#!/usr/bin/env ruby
require 'open3'

MARKER = "!XMP#{Time.new.to_i}_#{rand(1000000)}!"
XMP_RE = Regexp.new("^" + Regexp.escape(MARKER) + '\[([0-9]+)\] (=>|~>) (.*)')
VAR = "_xmp_#{Time.new.to_i}_#{rand(1000000)}"
WARNING_RE = /-:([0-9]+): warning: (.*)/

if ARGV[0] == "-unittest"
  mode = :unittest
  ARGV.shift
else
  mode = :normal
end

interpreter = ARGV.shift || "ruby"
code = ARGF.read

idx = 0
newcode = code.gsub(/^(.*) # =>.*/) do |l|
  expr = $1
  (/^\s*#/ =~ l) ? l :
  basic_eval = %!((#{VAR} = (#{expr}); $stderr.puts("#{MARKER}[#{idx+=1}] => " + #{VAR}.inspect) || #{VAR}))!
  case mode
  when :unittest
    %|begin; #{basic_eval}; rescue Exception; $stderr.puts("#{MARKER}[#{idx}] ~> " + $!.class.to_s); end|
  else
    basic_eval
  end
end
stdin, stdout, stderr = Open3::popen3(interpreter, "-w")
stdin.puts newcode
stdin.close
output = stderr.readlines

results = Hash.new{|h,k| h[k] = []}
exceptions = Hash.new{|h,k| h[k] = []}
output.grep(XMP_RE).each do |line|
  result_id, op, result = XMP_RE.match(line).captures
  case op
  when "=>"
    results[result_id.to_i] << result
  when "~>"
    exceptions[result_id.to_i] << result
  end
end

assertion = lambda do |expression, index|
  if !(wanted = results[index]).empty? || !exceptions[index]
    wanted = wanted + ["fillme"]
    case wanted
    when %w[nil fillme]
      "assert_nil(#{expression})"
    else
      "assert_equal(#{wanted[0]}, #{expression})"
    end
  else
    "assert_raise(#{exceptions[index][0]}){#{expression}}"
  end
end

idx = 0
annotated = code.gsub(/^(.*) # =>.*/) do |l|
  expr = $1
  next l if /^\s*#/ =~ l
  case mode
  when :unittest
    indent =  /^\s*/.match(l)[0]
    "#{indent}#{assertion[expr.strip, idx+=1]}"
  else
    "#{expr} # => " + (results[idx+=1] || []).join(", ")
  end
end.gsub(/ # !>.*/, '').gsub(/# (>>|~>)[^\n]*\n/m, "");

warnings = {}
output.join.grep(WARNING_RE).map do |x|
  md = WARNING_RE.match(x)
  warnings[md[1].to_i] = md[2]
end
idx = 0
annotated = annotated.map do |line|
  w = warnings[idx+=1]
  w ? (line.chomp + " # !> #{w}") : line
end
puts annotated
output.reject!{|x| /^-:[0-9]+: warning/.match(x)}
if exception = /^-:[0-9]+:.*/m.match(output.join)
  puts exception[0].map{|line| "# ~> " + line }
end
if mode == :normal && (s = stdout.read) != ""
  puts "# >> #{s.inspect}" 
end
